# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Qiskit Experiments implementation of GEM experiment protocol."""

from __future__ import annotations

from itertools import product
from typing import Any

import numpy as np

from qiskit_experiments.framework import BaseExperiment, Options
from qiskit.providers.backend import BackendV2
from qiskit.circuit import QuantumCircuit, Parameter

from .analysis import GemAnalysis
from .plaquettes import PlaquetteLattice


class GemExperiment(BaseExperiment):
    """GEM experiment in 2D protocol."""

    parameter = Parameter("θ")

    def __init__(
        self,
        plaquettes: PlaquetteLattice | list[int],
        backend: BackendV2 | None,
    ):
        """Create new GEM experiment.

        .. notes::
            Current implementation supports only Heavy Hexagon lattice
            and measurement-based 2D protocol.

        Args:
            plaquettes: List of plaquette indices or configuired :class:`PlaquetteLattice`
                instance to build GEM circuits on.
            backend: Qiskit Backend to run experiments.

        Raises:
            RuntimeError: When plaquettes are selected by index but backend is not provided.
        """
        if not isinstance(plaquettes, PlaquetteLattice):
            plaquettes = PlaquetteLattice.from_backend(backend).filter(includes=plaquettes)
        qubits = [q.index for q in plaquettes.qubits()]
        super().__init__(
            physical_qubits=qubits,
            analysis=GemAnalysis(plaquettes=plaquettes),
            backend=backend,
        )
        self._plaquettes = plaquettes

    @property
    def plaquettes(self) -> PlaquetteLattice:
        """Plaquette lattice to run experiemnts on."""
        return self._plaquettes

    def parameters(self) -> np.ndarray:
        """Create numpy array of parameters scanned in this experiment.

        Returns:
            Angle of ZZ gate determining tA.
        """
        if self.experiment_options.angles is None:
            return np.linspace(
                self.experiment_options.min_angle,
                self.experiment_options.max_angle,
                self.experiment_options.num_angles,
            )
        return np.asarray(self.experiment_options.angles, dtype=float)

    @classmethod
    def _default_transpile_options(cls) -> Options:
        options = super()._default_transpile_options()
        options.optimization_level = 1
        return options

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default option values for GEM experiment.

        Experiment Options:
            schedule_idx (int): Select from 0 to 11 if any. If not provided
                the experiment builds circuits for all possible scheduling patterns
                and performs the analysis on averaged outcomes.
            min_angle (float): Minimum angle to scan.
            max_angle (float): Maximum angle to scan.
            num_angles (int): Number of angles to scan.
            angles (Sequence[float]): The list of angles that will be scanned in
                the experiment. If not set, then ``num_angles``
                evenly spaced delays between ``min_angle`` and ``max_angle``
                are used. If ``angles`` is set, these parameters are ignored.
        """
        options = super()._default_experiment_options()
        options.schedule_idx = None
        options.sweep_type = "A"

        options.min_angle = 0
        options.max_angle = np.pi / 2
        options.num_angles = 21
        options.angles = None

        options.set_validator("sweep_type", ["A", "B"])

        return options

    def parameterized_circuits(self) -> tuple[QuantumCircuit, ...]:
        """Return GEM circuit with unbound parameters.

        Returns:
            Parameterized circuits.
        """
        bond_idxs = [
            self.physical_qubits.index(q.index)
            for q in self._plaquettes.qubits()
            if q.role == "Bond"
        ]

        # Create virtual circuit
        # Virtual qubit index is based off of list index of self.physical_qubits
        if (usr_index := self.experiment_options.schedule_idx) is not None:
            sched_idx = [usr_index]
        else:
            sched_idx = list(range(12))
        circs = []
        for i in sched_idx:
            sched_iter = self._plaquettes.build_gate_schedule(i)
            circ = QuantumCircuit(self.num_qubits, self.num_qubits)
            circ.metadata = {"schedule_index": i}
            circ.h(range(self.num_qubits))
            for gate_group in sched_iter:
                circ.barrier()
                for gate in gate_group:
                    if gate.group == self.experiment_options.sweep_type:
                        angle = self.parameter
                    else:
                        angle = np.pi / 2
                    circ.rzz(
                        angle,
                        self.physical_qubits.index(gate.index0),
                        self.physical_qubits.index(gate.index1),
                    )
            circ.barrier()
            circ.h(bond_idxs)
            circ.barrier()
            circ.measure(range(self.num_qubits), range(self.num_qubits))
            circs.append(circ)
        return tuple(circs)

    def circuits(self) -> list[QuantumCircuit]:
        circs_bound = []
        for tmp_circ, param in product(self.parameterized_circuits(), self.parameters()):
            assigned = tmp_circ.assign_parameters({self.parameter: param}, inplace=False)
            assigned.metadata["theta"] = np.round(param / np.pi, 5)
            circs_bound.append(assigned)
        return circs_bound

    def _metadata(self) -> dict[str, Any]:
        metadata = super()._metadata()
        metadata["connectivity"] = self.plaquettes.connectivity()
        metadata["plaquette_qubit_map"] = {p.index: p.qubits for p in self.plaquettes.plaquettes()}
        return metadata



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Plaquette representation of the Qiskit Backend."""

from __future__ import annotations

from typing import Iterator, Callable, NamedTuple

import numpy as np
from pymatching import Matching
from scipy.sparse import csc_matrix
from qiskit_experiments.framework import FigureData
from qiskit.providers import BackendV2

from gem_suite.gem_core import PyHeavyHexLattice, PyQubit, PyPlaquette, PyScheduledGate
from .plot_utils import dot_to_mplfigure


class ScheduledGate(NamedTuple):
    q0: int
    q1: int
    group: str


class DecodeOutcome(NamedTuple):
    w_ops: list[float]
    zxz_ops: list[float]
    f: tuple[float, float]
    g: tuple[float, float]


class PlaquetteLattice:
    """Plaquette representation of Qiskit Backend."""

    def __init__(self, lattice: PyHeavyHexLattice):
        """Create new plaquette lattice.

        Args:
            lattice: Rust plaquette core.
        """
        self._core = lattice

    @classmethod
    def from_backend(cls, backend: BackendV2):
        """Create new instance from Qiskit Backend."""
        if hasattr(backend, "configuration"):
            cmap = [tuple(qs) for qs in backend.configuration().coupling_map]
        else:
            cmap = list(backend.coupling_map)
        return PlaquetteLattice(PyHeavyHexLattice(cmap))

    @classmethod
    def from_coupling_map(cls, coupling_map: list[tuple[int, int]]):
        """Create new instance from device coupling map."""
        return PlaquetteLattice(PyHeavyHexLattice([tuple(qs) for qs in coupling_map]))

    def qubits(self) -> Iterator[PyQubit]:
        """Yield annotated qubit dataclasses."""
        yield from self._core.qubits()

    def plaquettes(self) -> Iterator[PyPlaquette]:
        """Yield plaquette dataclasses."""
        yield from self._core.plaquettes()

    def connectivity(self) -> list[tuple[int, int]]:
        """Return unidirectional qubit connectivity."""
        return self._core.connectivity()

    def draw_qubits(self) -> FigureData:
        """Draw coupling graph with qubits in the lattice."""
        svg_fig = dot_to_mplfigure(self._core.qubit_graph_dot(), "fdp", 300)
        return FigureData(svg_fig, name="qubit_graph")

    def draw_plaquettes(self) -> FigureData:
        """Draw coupling graph with plaquette in the lattice."""
        svg_fig = dot_to_mplfigure(self._core.plaquette_graph_dot(), "neato", 300)
        return FigureData(svg_fig, name="plaquette_graph")

    def draw_decode_graph(self) -> FigureData:
        """Draw qubit graph with annotation for decoding."""
        svg_fig = dot_to_mplfigure(self._core.decode_graph_dot(), "fdp", 300)
        return FigureData(svg_fig, name="decode_graph")

    def filter(self, includes: list[int]) -> PlaquetteLattice:
        """Create new plaquette lattice instance with subset of plaquettes.

        Args:
            includes: Index of plaquettes to include.
                This cannot include disconnected groups.
                All qubits must be connected to build GEM circuit.

        Returns:
            New plaquette lattice instance.
        """
        return PlaquetteLattice(self._core.filter(includes))

    def build_gate_schedule(self, index: int) -> Iterator[list[PyScheduledGate]]:
        """Yield list of entangling gates that can be simultaneously applied.

        Args:
            index: Index of gate schedule. There might be multiple scheduling patterns.

        Yields:
            List of :class:`.PyScheduledGate` dataclass representing
                an entangling gate, and all gates in a list can be
                applied simultaneously without qubit overlapping.
        """
        yield from self._core.build_gate_schedule(index)

    def check_matrix_csc(self) -> csc_matrix:
        """Create check matrix from the plaquette lattice."""
        hvec, dims = self._core.check_matrix_csc()
        data = len(hvec[0])
        return csc_matrix((data, hvec), dims, dtype=bool)

    def decode_outcomes(
        self,
        counts: dict[str, int],
        decoder: str = "pymatching",
        return_counts: bool = False,
    ) -> DecodeOutcome | tuple[DecodeOutcome, dict[str, int]]:
        """Decode count dictionary of the experiment result and analyze.

        Args:
            counts: Count dictionary of single circuit.
            decoder: MWPM decoder to decode count string. Either "pymatching" or "fusion-blossom".
            return_counts: Set True to return count dictionary.

        Returns:
            Decoded outcomes including plaquette and ZXZ bond operators,
            and f and g quantities associated with the prepared state magnetization.
            When the return_counts is set, this returns a tuple of
            outcome and count dictionary keyed on decoded site qubit bitstring.
        """
        if decoder == "fusion-blossom":
            out = self._core.decode_outcomes_fb(counts, return_counts)
            outcome = DecodeOutcome(*out[1:])
        elif decoder == "pymatching":
            hvec, dims = self._core.check_matrix_csc()
            data = np.ones(len(hvec[0]))
            solver = _build_batch_decoder(csc_matrix((data, hvec), dims, dtype=np.bool_))
            out = self._core.decode_outcomes_pm(solver, counts, return_counts)
            outcome = DecodeOutcome(*out[1:])
        else:
            raise ValueError(f"MWPM decoder {decoder} is not supported.")
        if return_counts:
            return outcome, out[0]
        return outcome


def _build_batch_decoder(check_matrix) -> Callable:

    solver = Matching.from_check_matrix(check_matrix)

    def _batched_solver(shots: list[bool], shape: tuple[int, int]):
        batch_in = np.array(shots, dtype=np.bool_).reshape(shape)
        out = solver.decode_batch(shots=batch_in)
        return out.astype(np.bool_).flatten().tolist()

    return _batched_solver



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Analysis for prepared state magnetization."""
from __future__ import annotations

import warnings
from typing import Callable

import pandas as pd
import numpy as np

from scipy.optimize import least_squares
from uncertainties import ufloat, UFloat, unumpy as unp, correlated_values
from qiskit_experiments.framework import AnalysisResultData, FigureData
from qiskit_experiments.database_service.device_component import Qubit

from gem_suite.plot_utils import plot_schedules
from gem_suite.gem_core import PyQubit, PyPlaquette


# pylint: disable=too-many-locals
def analyze_operators(
    data: pd.DataFrame,
) -> tuple[list[FigureData], list[AnalysisResultData]]:
    """Analyze plaquette and bond operators.

    .. notes::
        This analysis is performed on mean quantities
        averaged over all components, i.e. plaquettes and bonds.

    Args:
        data: Dataframe contaning w and zxz entries.

    Returns:
        FigureData and AnalysisResultData for Qiskit Experiments.
    """
    figures = []
    analysis_results = []

    # Average over all components
    w_op_data_mean = (
        data[data.name == "w"]
        .groupby(
            ["theta", "schedule"],
            as_index=False,
        )
        .value.agg(lambda v: ufloat(np.average(v), np.std(v)))
    )
    zxz_op_data_mean = (
        data[data.name == "zxz"]
        .groupby(
            ["theta", "schedule"],
            as_index=False,
        )
        .value.agg(lambda v: ufloat(np.average(v), np.std(v)))
    )

    # Fit data
    # Here we consider multi-objective fitting for both plaquette and bond models
    # since they share the p_bond parameter describing the
    # readout error on bond qubits and noise during entangling process.
    ax_w_op, (xvals, yvals_w) = plot_schedules(w_op_data_mean)
    ax_w_op.set_ylabel(r"Mean plaquette observables $\langle W \rangle$")

    ax_zxz_op, (_, yvals_zxz) = plot_schedules(zxz_op_data_mean)
    ax_zxz_op.set_ylabel(r"Mean bond observables $\langle ZXZ \rangle$")

    yvals_w_nominal = unp.nominal_values(yvals_w)
    yvals_w_std = unp.std_devs(yvals_w)
    yvals_w_std = np.where(np.isclose(yvals_w_std, 0.0), np.finfo(float).eps, yvals_w_std)
    w_weights = 1 / yvals_w_std
    w_weights = np.clip(w_weights, 0.0, np.percentile(w_weights, 90))

    yvals_zxz_nominal = unp.nominal_values(yvals_zxz)
    yvals_zxz_std = unp.std_devs(yvals_zxz)
    yvals_zxz_std = np.where(np.isclose(yvals_zxz_std, 0.0), np.finfo(float).eps, yvals_zxz_std)
    zxz_weights = 1 / yvals_zxz_std
    zxz_weights = np.clip(zxz_weights, 0.0, np.percentile(zxz_weights, 90))

    def multi_residuals(params):
        r_w = (w_model(xvals, params[0]) - yvals_w_nominal) * w_weights
        r_zxz = (zxz_model(xvals, params[0], params[1]) - yvals_zxz_nominal) * zxz_weights
        return np.concatenate([r_w, r_zxz])

    if ret := fit_util(
        multi_residuals,
        init=[0.01, 0.01],
        bounds=([0.0, 0.0], [1.0, 1.0]),
        n_data=len(xvals),
    ):
        fit_vals, red_chisq, aic = ret

        # Plot fit curves
        x_interp = np.linspace(0, 0.5, 100)
        y_interp_w = w_model(x_interp, fit_vals[0].n)
        y_interp_zxz = zxz_model(x_interp, fit_vals[0].n, fit_vals[1].n)
        ax_w_op.plot(
            x_interp,
            y_interp_w,
            color="blue",
            lw=2.0,
        )
        ax_w_op.text(
            x_interp[50],
            y_interp_w[50],
            f"p_bond = {fit_vals[0].n:.4f}",
            ha="left",
            va="top",
        )
        ax_zxz_op.plot(
            x_interp,
            y_interp_zxz,
            color="blue",
            lw=2.0,
        )
        ax_zxz_op.text(
            x_interp[50],
            y_interp_zxz[50],
            f"p_bond = {fit_vals[0].n:.4f}, p_site = {fit_vals[1].n:.4f}",
            ha="left",
            va="top",
        )

        # Create results
        analysis_results = [
            AnalysisResultData(
                "p_bond_mean",
                value=fit_vals[0],
                chisq=red_chisq,
                quality="Good" if red_chisq < 3.0 else "Bad",
                extra={"aic": aic},
            ),
            AnalysisResultData(
                "p_site_mean",
                value=fit_vals[1],
                chisq=red_chisq,
                quality="Good" if red_chisq < 3.0 else "Bad",
                extra={"aic": aic},
            ),
        ]
    else:
        warnings.warn(
            "Fitting for global p_bond and p_site is not successful. ",
            RuntimeWarning,
        )
        # Create results
        analysis_results = [
            AnalysisResultData(
                "p_bond_mean",
                value=None,
                quality="Bad",
            ),
            AnalysisResultData(
                "p_site_mean",
                value=None,
                quality="Bad",
            ),
        ]
    figures.extend(
        [
            FigureData(
                figure=ax_w_op.get_figure(),
                name="plaquette_ops",
            ),
            FigureData(
                figure=ax_zxz_op.get_figure(),
                name="bond_ops",
            ),
        ]
    )

    return figures, analysis_results


def analyze_individual_bonds(
    data: pd.DataFrame,
    qubits: list[PyQubit],
) -> list[AnalysisResultData]:
    """Analyze bond operators individually.

    Args:
        data: Dataframe contaning zxz entries.

    Returns:
        AnalysisResultData for Qiskit Experiments.
    """
    analysis_results = []

    # Create map from bond index to qubit
    bond_qubits = sorted([q for q in qubits if q.role == "Bond"], key=lambda q: q.index)

    # ZXZ data averaged over all schedules
    zxz_op_data = (
        data[data.name == "zxz"]
        .groupby(["theta", "component"], as_index=False)
        .value.agg(lambda v: ufloat(np.average(v), np.std(v)))
    )
    bonds = sorted(zxz_op_data.component.unique())
    for bond, qubit in zip(bonds, bond_qubits):
        qk_components = [
            Qubit(qubit.index),
            Qubit(qubit.neighbors[0]),
            Qubit(qubit.neighbors[1]),
        ]
        subdata = zxz_op_data[zxz_op_data.component == bond]
        xvals = subdata.theta.to_numpy()
        yvals = subdata.value.to_numpy()
        yvals_nominal = unp.nominal_values(yvals)
        yvals_std = unp.std_devs(yvals)
        yvals_std = np.where(np.isclose(yvals_std, 0.0), np.finfo(float).eps, yvals_std)
        weights = 1 / yvals_std
        weights = np.clip(weights, 0.0, np.percentile(weights, 90))

        # pylint: disable=cell-var-from-loop
        def objective(params):
            return zxz_model(xvals, params[0], params[1] - yvals_nominal) * weights

        if ret := fit_util(
            objective,
            init=[0.01, 0.01],
            bounds=([0.0, 0.0], [1.0, 1.0]),
            n_data=len(xvals),
        ):
            fit_vals, red_chisq, aic = ret
            # Plotting for every bond generates too many figures.
            # Just create results entries.
            this_results = [
                AnalysisResultData(
                    "p_bond",
                    value=fit_vals[0],
                    chisq=red_chisq,
                    quality="Good" if red_chisq < 3.0 else "Bad",
                    device_components=qk_components,
                    extra={"aic": aic},
                ),
                AnalysisResultData(
                    "p_site",
                    value=fit_vals[1],
                    chisq=red_chisq,
                    quality="Good" if red_chisq < 3.0 else "Bad",
                    device_components=qk_components,
                    extra={"aic": aic},
                ),
            ]
        else:
            warnings.warn(
                f"Fitting for p_bond and p_site on bond {bond} is not successful. ",
                RuntimeWarning,
            )
            this_results = [
                AnalysisResultData(
                    "p_bond",
                    value=None,
                    quality="Bad",
                    device_components=qk_components,
                ),
                AnalysisResultData(
                    "p_site",
                    value=None,
                    quality="Bad",
                    device_components=qk_components,
                ),
            ]
        analysis_results.extend(this_results)
    return analysis_results


def analyze_clifford_limit(
    data: pd.DataFrame,
    plaquettes: list[PyPlaquette],
    qubits: list[PyQubit],
) -> list[AnalysisResultData]:
    """Visualize quality distribution at Clifford limit."""
    analysis_results = []

    clif_data = data[np.isclose(data.theta - 0.5, np.finfo(float).eps)]
    if len(clif_data) == 0:
        warnings.warn(
            "Experiment data at theta = 0.5 pi of the Clifford "
            "limit is not included in the result.",
            RuntimeWarning,
        )
        return []
    plaquettes = sorted(plaquettes, key=lambda p: p.index)
    bond_to_qubits = dict(
        enumerate(sorted([q for q in qubits if q.role == "Bond"], key=lambda q: q.index))
    )
    # pylint: disable=invalid-name
    for wi, plaquette in enumerate(plaquettes):
        all_obs = []
        sub_qubits = plaquette.qubits
        plq_index = plaquette.index
        filtered_data = clif_data[(clif_data.name == "w") & (clif_data.component == wi)]
        w_val = ufloat(np.average(filtered_data.value), np.std(filtered_data.value))
        all_obs.append(w_val)
        for bond_bit, bond_qubit in bond_to_qubits.items():
            # Take bond qubits in this plaquette
            if bond_qubit.index not in sub_qubits:
                continue
            filtered_data = clif_data[(clif_data.name == "zxz") & (clif_data.component == bond_bit)]
            zxz_val = ufloat(np.average(filtered_data.value), np.std(filtered_data.value))
            all_obs.append(zxz_val)
        if any(o < 0 for o in all_obs):
            value = None
        else:
            value = np.prod(all_obs) ** (1 / len(all_obs))
        analysis_results.append(
            AnalysisResultData(
                "plaquette_quality",
                value=value,
                device_components=[Qubit(q) for q in sub_qubits],
                extra={"plaquette": plq_index},
            )
        )
    return analysis_results


def fit_util(
    objective: Callable,
    init: list[float],
    bounds: tuple[list[float], list[float]],
    n_data: int,
) -> None | tuple[tuple[UFloat, ...], float, float]:
    """Wrapper of scipy optmizer function.

    Args:
        objective: An objective function to minimize.
        init: Initial parameters.
        bounds: Minimum and maximum range of parameters.
        n_data: Number of data points to fit.

    Returns:
        Parameters in ufloat, reduced Chi-squared, and AIC when successful.
    """
    if n_data < len(init):
        warnings.warn(
            "Number of experiment data points is smaller than free parameters to fit. "
            "Fit cannot be performed on this experiment result."
        )
        return None
    ret = least_squares(
        fun=objective,
        x0=init,
        bounds=bounds,
        method="trf",
        loss="linear",
    )
    if ret.success:
        # Compute statistics
        dof = n_data - len(init)
        chisq = np.sum(ret.fun**2)
        red_chisq = chisq / dof
        if chisq != 0:
            aic = n_data * np.log(chisq / n_data) + 2 * len(init)
        else:
            aic = None
        # Compute error of the parameters
        hess = np.matmul(ret.jac.T, ret.jac)
        fit_vals = None
        try:
            covar = np.linalg.inv(hess)
            if not any(np.diag(covar) < 0):
                fit_vals = correlated_values(ret.x, covariance_mat=covar)
        except np.linalg.LinAlgError:
            pass
        return (fit_vals or tuple(ufloat(x, np.nan) for x in ret.x), red_chisq, aic)
    return None


def w_model(theta, p_bond):
    """Fit model for plaquette observable.

    Args:
        theta: Experiment parameter.
        p_bond: Readout error on bond qubits and noise during entangling process.
    """
    return np.sin(np.pi * theta) ** 6 * (1 - 2 * p_bond) ** 6


def zxz_model(theta, p_bond, p_site):
    """Fit model for bond observable.

    Args:
        theta: Experiment parameter.
        p_bond: Readout error on bond qubits and noise during entangling process.
        p_site: Readout error on site qubits.
    """
    return np.sin(np.pi * theta) * (1 - 2 * p_site) ** 2 * (1 - 2 * p_bond)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Sub module for visualization."""
from __future__ import annotations

import subprocess
import tempfile
import io

from typing import cast, TYPE_CHECKING

import pandas as pd
import numpy as np
from uncertainties import unumpy as unp
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from qiskit_experiments.framework.matplotlib import get_non_gui_ax

try:
    from PIL import Image  # type: ignore

    HAS_PILLOW = True
except ImportError:
    HAS_PILLOW = False

if TYPE_CHECKING:
    from PIL import Image  # type: ignore


def plot_schedules(
    data: pd.DataFrame,
) -> tuple[Axes, tuple[np.ndarray, np.ndarray]]:
    """Plot individual data and average of them in dataframe.

    Args:
        data: Filtered dataframe containing data to visualize.

    Returns:
        Matplotlib figure axis object and averaged data in ufloat format.
    """
    axis: Axes = get_non_gui_ax()
    data = data.sort_values("theta")
    schedules = sorted(data.schedule.unique())
    if len(schedules) > 1:
        for sched in schedules:
            subdata = data[data.schedule == sched]
            nominal = unp.nominal_values(subdata.value)
            stdev = unp.std_devs(subdata.value)
            axis.plot(
                subdata.theta,
                nominal,
                color="grey",
                alpha=0.4,
                lw=0.7,
            )
            axis.fill_between(
                subdata.theta,
                nominal + stdev,
                nominal - stdev,
                color="grey",
                alpha=0.1,
            )
        avg = data.groupby("theta", as_index=False).value.agg(np.average)
        x_mean = avg.theta.to_numpy()
        y_mean = avg.value.to_numpy()
    else:
        x_mean = data.theta.to_numpy()
        y_mean = data.value.to_numpy()
    nominal = unp.nominal_values(y_mean)
    stdev = unp.std_devs(y_mean)
    axis.errorbar(
        x_mean,
        nominal,
        color="blue",
        alpha=1.0,
        fmt="o",
    )
    axis.set_xlabel(r"$\theta$ ($\pi$)")
    axis.set_xlim(0, max(0.5, x_mean.max()))
    return axis, (x_mean.astype(float), y_mean)


def dot_to_mplfigure(
    dot_data: str,
    method: str,
    dpi: int = 300,
    rescale: int = 4,
) -> Figure:
    """Render dot script and return in matplotlib Figure format.

    Args:
        dot_data: Input dot script.
        method: Drawing method.
        dpi: Dot per inch.
        rescale: Scaling factor of image to load PIL Image data in the matplotlib canvas.

    Returns:
        Matplotlib Figure data written in SVG Backend.
        This backend doesn't support automatic rendering in Jupyter notebook environment.
    """
    if not HAS_PILLOW:
        raise ImportError(
            "Pillow is necessary to use draw(). "
            "It can be installed with 'pip install pydot pillow'"
        )
    try:
        subprocess.run(
            ["dot", "-V"],
            cwd=tempfile.gettempdir(),
            check=True,
            capture_output=True,
        )
    except Exception as ex:
        raise RuntimeError(
            "Graphviz could not be found or run. "
            "This function requires that Graphviz is installed."
        ) from ex
    dot_result = subprocess.run(
        [method, "-T", "png"],
        input=cast(str, dot_data).encode("utf-8"),
        capture_output=True,
        encoding=None,
        check=True,
        text=False,
    )
    dot_bytes_image = io.BytesIO(dot_result.stdout)
    img = Image.open(dot_bytes_image)

    # Convert into matplotlib axis object
    axis: Axes = get_non_gui_ax()
    fig = axis.get_figure()
    fig.set_size_inches(img.width / dpi * rescale, img.height / dpi * rescale)
    fig.dpi = dpi

    axis.imshow(img)
    axis.set_axis_off()
    return fig



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Callable

def visualize_plaquette_with_noise(
    plaquette_qubits_map: dict[int, list[int]],
    noise_map: dict[int, float],
) -> str: ...

class PyQubit:
    index: int
    role: str
    group: str
    coordinate: tuple[int, int] | None
    neighbors: list[int]

class PyPlaquette:
    index: int
    qubits: list[int]
    neighbors: list[int]

class PyScheduledGate:
    index0: int
    index1: int
    group: str

class PyHeavyHexLattice:
    def __init__(self, coupling_map: list[tuple[int, int]]) -> PyHeavyHexLattice: ...
    @classmethod
    def from_plaquettes(
        cls,
        plaquette_qubits_map: dict[int, list[int]],
        connectivity: list[tuple[int, int]],
    ) -> PyHeavyHexLattice: ...
    def qubit_graph_dot(self) -> str: ...
    def plaquette_graph_dot(self) -> str: ...
    def decode_graph_dot(self) -> str: ...
    def qubits(self) -> list[int]: ...
    def plaquettes(self) -> list[int]: ...
    def connectivity(self) -> list[tuple[int, int]]: ...
    def filter(self, includes: list[int]) -> PyHeavyHexLattice: ...
    def build_gate_schedule(self, index: int) -> list[list[PyScheduledGate]]: ...
    def decode_outcomes_fb(
        self, counts: dict[str, int], return_counts: bool
    ) -> tuple[
        dict[str, int] | None,
        list[float],
        list[float],
        tuple[float, float],
        tuple[float, float],
    ]: ...
    def decode_outcomes_pm(
        self, solver: Callable, counts: dict[str, int], return_counts: bool
    ) -> tuple[
        dict[str, int] | None,
        list[float],
        list[float],
        tuple[float, float],
        tuple[float, float],
    ]: ...
    def check_matrix_csc(
        self,
    ) -> tuple[tuple[list[int], list[int]], tuple[float, float]]: ...



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Qiskit Experiments implementation of GEM outcome decoding and analysis."""

from __future__ import annotations

import pandas as pd
from uncertainties import ufloat
from qiskit_experiments.framework import (
    BaseAnalysis,
    ExperimentData,
    AnalysisResultData,
    ArtifactData,
    FigureData,
    Options,
)

from gem_suite.gem_core import visualize_plaquette_with_noise, PyHeavyHexLattice
from .sub_analysis import (
    analyze_magnetization,
    analyze_operators,
    analyze_individual_bonds,
    analyze_clifford_limit,
)
from .plaquettes import PlaquetteLattice
from .plot_utils import dot_to_mplfigure


class GemAnalysis(BaseAnalysis):
    """GEM analysis for 2D protocol."""

    def __init__(
        self,
        plaquettes: PlaquetteLattice | None = None,
    ):
        """Create new GEM analysis instance.

        Args:
            plaquettes: PlaquetteLattice instance of target device.
                This can be automatically instantiated from the
                GEM Experiment metadata if not given.
        """
        super().__init__()
        self._plaquettes = plaquettes

    @classmethod
    def _default_options(cls) -> Options:
        """Default analysis options for GEM analysis.

        Analysis Options:
            analyze_individual_bond (bool): Perform analysis on individual
                ZXZ operator of each bond qubits to analyze
                local p_bond and p_site quantities.
                This may create large amount of analysis result entries
                when you run experiment with many plaquettes.
            analyze_clifford_limit (bool): Perform analysis at Clifford limit,
                i.e. theta = 0.5 pi. This data point must be included.
                Enabling this option will add a new figure illustrating
                the error distribution in the plaquette view along with
                several analysis results of plaquette-wise qualities.
            decoder (str): Minimum weight perfect matching decoder.
                "fusion-blossom" and "pymatching" are supported.
        """
        options = super()._default_options()
        options.update_options(
            analyze_individual_bond=False,
            analyze_clifford_limit=False,
            decoder="pymatching",
        )
        options.set_validator("decoder", ["pymatching", "fusion-blossom"])
        return options

    # pylint: disable=too-many-locals
    def _run_analysis(
        self,
        experiment_data: ExperimentData,
    ) -> tuple[list[AnalysisResultData | ArtifactData], list[FigureData]]:
        fig_data = []
        analysis_results = []

        # Check if lattice object exists.
        # The lattice information is necessary for decoding experiment outcomes.
        if self._plaquettes is None:
            plaquette_qubit_map = experiment_data.metadata.get("plaquette_qubit_map", None)
            connectivity = experiment_data.metadata.get("connectivity", None)
            if plaquette_qubit_map is None or connectivity is None:
                raise RuntimeError(
                    "Analysis is not initialized with PlaquetteLattice instance and "
                    "its description is not provided by the experiment metadata. "
                    "This analysis cannot be performed because of "
                    "the missing lattice information to decode bitstrings."
                )
            lattice = PyHeavyHexLattice.from_plaquettes(plaquette_qubit_map, connectivity)
            self._plaquettes = PlaquetteLattice(lattice)

        # Decode count dictionaries and build intermediate data collection.
        records = []
        for data in experiment_data.data():
            result = self._plaquettes.decode_outcomes(
                counts=data["counts"],
                decoder=self.options.decoder,
                return_counts=False,
            )
            records.extend(
                [
                    {
                        "name": "f",
                        "theta": data["metadata"]["theta"],
                        "schedule": data["metadata"]["schedule_index"],
                        "value": ufloat(result.f[0], result.f[1]),
                        "component": pd.NA,
                    },
                    {
                        "name": "g",
                        "theta": data["metadata"]["theta"],
                        "schedule": data["metadata"]["schedule_index"],
                        "value": ufloat(result.g[0], result.g[1]),
                        "component": pd.NA,
                    },
                ]
            )
            for i, expv in enumerate(result.w_ops):
                records.append(
                    {
                        "name": "w",
                        "theta": data["metadata"]["theta"],
                        "schedule": data["metadata"]["schedule_index"],
                        "value": expv,
                        "component": i,
                    }
                )
            for i, expv in enumerate(result.zxz_ops):
                records.append(
                    {
                        "name": "zxz",
                        "theta": data["metadata"]["theta"],
                        "schedule": data["metadata"]["schedule_index"],
                        "value": expv,
                        "component": i,
                    }
                )
        dataframe = pd.DataFrame.from_records(records)
        del records

        # Analyze magnetizations
        num_site = 0
        for qubit in self._plaquettes.qubits():
            if qubit.role == "Site":
                num_site += 1
        mag_figs, mag_results = analyze_magnetization(dataframe, num_site)
        fig_data.extend(mag_figs)
        analysis_results.extend(mag_results)

        # Analyze operators
        op_figs, op_results = analyze_operators(dataframe)
        fig_data.extend(op_figs)
        analysis_results.extend(op_results)

        # Analyze local bond parameters (Optional)
        if self.options.analyze_individual_bond:
            local_params = analyze_individual_bonds(
                data=dataframe,
                qubits=self._plaquettes.qubits(),
            )
            analysis_results.extend(local_params)

        # Analyze plaquette quality (Optional)
        if self.options.analyze_clifford_limit:
            plaquette_qualities = analyze_clifford_limit(
                data=dataframe,
                plaquettes=self._plaquettes.plaquettes(),
                qubits=self._plaquettes.qubits(),
            )
            analysis_results.extend(plaquette_qualities)
            if plaquette_qualities:
                p_map = {p.index: p.qubits for p in self._plaquettes.plaquettes()}
                noise_map = {
                    q.extra["plaquette"]: q.value.n
                    for q in plaquette_qualities
                    if q.value is not None
                }
                dot = visualize_plaquette_with_noise(p_map, noise_map)
                fig_data.append(
                    FigureData(dot_to_mplfigure(dot, "neato", 300), name="plaquette_quality")
                )

        return analysis_results, fig_data



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Analysis for prepared state magnetization."""
from __future__ import annotations

import pandas as pd
import numpy as np

from uncertainties import unumpy as unp
from qiskit_experiments.framework import AnalysisResultData, FigureData

from gem_suite.plot_utils import plot_schedules


def analyze_magnetization(
    data: pd.DataFrame,
    num_sites: int,
) -> tuple[list[FigureData], list[AnalysisResultData]]:
    """Analyze magnetization of the prepared state.

    Args:
        data: Dataframe contaning f and g entries.

    Returns:
        FigureData and AnalysisResultData for Qiskit Experiments.
    """
    figures = []
    analysis_results = []

    # Analyze two point classical correlation
    f_data = data[data.name == "f"]
    f_data.loc[:, "value"] /= num_sites
    axis, (xvals, yvals) = plot_schedules(f_data)
    nominal_yvals = unp.nominal_values(yvals)
    argmax = np.argmax(nominal_yvals)
    fmax = nominal_yvals[argmax]
    axis.set_ylabel(r"Average two-point correlation $f$")
    axis.text(
        xvals[argmax],
        nominal_yvals[argmax],
        f"max_tpc = {fmax:.4f}",
        ha="right",
        va="bottom",
    )
    figures.append(
        FigureData(
            figure=axis.get_figure(),
            name="two_point_correlation",
        )
    )
    analysis_results.append(
        AnalysisResultData(
            name="max_tpc",
            value=yvals[argmax],
        )
    )

    # Analyze variance
    g_data = data[data.name == "g"]
    axis, (xvals, yvals) = plot_schedules(g_data)
    nominal_yvals = unp.nominal_values(yvals)
    argmax = np.argmax(nominal_yvals)
    critical = xvals[argmax]
    axis.set_ylabel(r"Normalized variance $g$")
    axis.text(
        critical,
        nominal_yvals[argmax],
        f"critical_angle = {critical:.4f}",
        ha="right",
        va="bottom",
    )
    figures.append(
        FigureData(
            figure=axis.get_figure(),
            name="normalized_variance",
        )
    )
    analysis_results.append(
        AnalysisResultData(
            name="critical_angle",
            value=critical,
        )
    )

    return figures, analysis_results



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""GEM Analysis test."""

import unittest

import numpy as np
from gem_suite.plaquettes import PlaquetteLattice

from .utils import process_counts_debug


class TestAnalysis(unittest.TestCase):
    """Test case for GEM analysis with Falcon processor mock."""

    def setUp(self):
        super().setUp()

        falcon_cmap = [
            (0, 1),
            (1, 0),
            (1, 2),
            (1, 4),
            (2, 1),
            (2, 3),
            (3, 2),
            (3, 5),
            (4, 1),
            (4, 7),
            (5, 3),
            (5, 8),
            (6, 7),
            (7, 4),
            (7, 6),
            (7, 10),
            (8, 5),
            (8, 9),
            (8, 11),
            (9, 8),
            (10, 7),
            (10, 12),
            (11, 8),
            (11, 14),
            (12, 10),
            (12, 13),
            (12, 15),
            (13, 12),
            (13, 14),
            (14, 11),
            (14, 13),
            (14, 16),
            (15, 12),
            (15, 18),
            (16, 14),
            (16, 19),
            (17, 18),
            (18, 15),
            (18, 17),
            (18, 21),
            (19, 16),
            (19, 20),
            (19, 22),
            (20, 19),
            (21, 18),
            (21, 23),
            (22, 19),
            (22, 25),
            (23, 21),
            (23, 24),
            (24, 23),
            (24, 25),
            (25, 22),
            (25, 24),
            (25, 26),
            (26, 25),
        ]

        self.falcon_lattice = PlaquetteLattice.from_coupling_map(falcon_cmap)

    def test_fusion_blossom_decode(self):
        """Test count decode with fusion blossom solver."""
        rng = np.random.default_rng(123)
        counts = {"".join(rng.choice(["0", "1"], size=21)): 100 for _ in range(10)}
        test_outcomes, test_dict = self.falcon_lattice.decode_outcomes(
            counts=counts,
            decoder="fusion-blossom",
            return_counts=True,
        )
        ref = process_counts_debug(counts)
        self.assertDictEqual(test_dict, ref[0])
        np.testing.assert_array_almost_equal(
            np.array(test_outcomes.w_ops, dtype=float),
            np.array(ref[1], dtype=float),
        )
        np.testing.assert_array_almost_equal(
            np.array(test_outcomes.zxz_ops, dtype=float),
            np.array(ref[2], dtype=float),
        )
        self.assertAlmostEqual(test_outcomes.f[0], ref[3][0])
        self.assertAlmostEqual(test_outcomes.f[1], ref[3][1])
        self.assertAlmostEqual(test_outcomes.g[0], ref[4][0])
        self.assertAlmostEqual(test_outcomes.g[1], ref[4][1])

    def test_pymatching_decode(self):
        """Test count decode with pymatching solver."""
        rng = np.random.default_rng(123)
        counts = {"".join(rng.choice(["0", "1"], size=21)): 100 for _ in range(10)}
        test_outcomes, test_dict = self.falcon_lattice.decode_outcomes(
            counts=counts,
            decoder="pymatching",
            return_counts=True,
        )
        ref = process_counts_debug(counts)
        self.assertDictEqual(test_dict, ref[0])
        np.testing.assert_array_almost_equal(
            np.array(test_outcomes.w_ops, dtype=float),
            np.array(ref[1], dtype=float),
        )
        np.testing.assert_array_almost_equal(
            np.array(test_outcomes.zxz_ops, dtype=float),
            np.array(ref[2], dtype=float),
        )
        self.assertAlmostEqual(test_outcomes.f[0], ref[3][0])
        self.assertAlmostEqual(test_outcomes.f[1], ref[3][1])
        self.assertAlmostEqual(test_outcomes.g[0], ref[4][0])
        self.assertAlmostEqual(test_outcomes.g[1], ref[4][1])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"GEM Suite setup file."

from setuptools import setup
from setuptools_rust import Binding, RustExtension

setup(
    rust_extensions=[
        RustExtension(
            "gem_suite.gem_core",
            binding=Binding.PyO3,
        )
    ],
    options={"bdist_wheel": {"py_limited_api": "cp38"}},
    long_description_content_type="text/markdown",
)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""GEM circuit test."""

import unittest
from math import pi

import numpy as np
from qiskit_ibm_runtime.fake_provider import FakeGeneva
from qiskit import QuantumCircuit

from gem_suite.experiments import GemExperiment


class TestCircuit(unittest.TestCase):
    """Test that generates GEM circuit."""

    def test_generate_circuit(self):
        """Test comparing virtual circuits."""
        exp = GemExperiment(range(2), backend=FakeGeneva())

        exp.set_experiment_options(schedule_idx=10)
        test_circ = exp.parameterized_circuits()[0]

        theta = exp.parameter

        ref_circ = QuantumCircuit(21, 21)
        ref_circ.h(range(21))
        ref_circ.barrier()
        ref_circ.rzz(theta, [4, 7, 13, 16], [6, 9, 15, 18])
        ref_circ.rzz(pi / 2, [3, 8, 12, 17], [5, 11, 14, 20])
        ref_circ.barrier()
        ref_circ.rzz(theta, [0, 6, 9, 15], [3, 8, 12, 17])
        ref_circ.rzz(pi / 2, [1, 10, 19], [2, 11, 20])
        ref_circ.barrier()
        ref_circ.rzz(theta, [0, 9, 18], [1, 10, 19])
        ref_circ.rzz(pi / 2, [2, 5, 11, 14], [4, 7, 13, 16])
        ref_circ.barrier()
        ref_circ.h([1, 3, 4, 7, 8, 10, 12, 13, 16, 17, 19])
        ref_circ.barrier()
        ref_circ.measure(range(21), range(21))

        self.assertEqual(test_circ, ref_circ)

    def test_qubit_layout(self):
        """Test physical qubit index list."""
        exp = GemExperiment(range(2), backend=FakeGeneva())
        self.assertListEqual(
            list(exp.physical_qubits),
            [
                1,
                2,
                3,
                4,
                5,
                7,
                8,
                10,
                11,
                12,
                13,
                14,
                15,
                16,
                18,
                19,
                21,
                22,
                23,
                24,
                25,
            ],
        )

    def test_parameter_list(self):
        """Test parameter (theta) list to scan."""
        exp = GemExperiment(range(2), backend=FakeGeneva())

        np.testing.assert_array_almost_equal(
            exp.parameters(),
            np.linspace(0, np.pi / 2, 21),
        )
